@@ -32,9 +32,12 @@ import time
 import urllib
 import zipfile
 try:
-    from cStringIO import StringIO
+    from io import BytesIO
 except ImportError:
-    from StringIO import StringIO
+    try:
+        from cStringIO import StringIO as BytesIO
+    except ImportError:
+        from StringIO import StringIO as BytesIO
 from copy import deepcopy
 import datetime
 from decimal import Decimal
@@ -130,7 +133,7 @@ class ImageHref:
         elif isinstance(bitstream, ChartTemplate):
             bitstream = bitstream.generate(**self.context).render()
         elif not hasattr(bitstream, 'seek') or not hasattr(bitstream, 'read'):
-            bitstream = StringIO(bitstream)
+            bitstream = BytesIO(bitstream)
         bitstream.seek(0)
         file_content = bitstream.read()
         name = md5(file_content).hexdigest()
@@ -237,6 +240,7 @@ class Template(MarkupTemplate):
         self.namespaces = {}
         self.inner_docs = []
         self.has_col_loop = False
+        self._zip_source = None
         super(Template, self).__init__(source, filepath, filename, loader,
                                        encoding, lookup, allow_exec)
 
@@ -245,7 +249,19 @@ class Template(MarkupTemplate):
 
         It adds genshi directives and finds the inner docs.
         """
-        zf = zipfile.ZipFile(self.filepath)
+        if not self.filepath:
+            if hasattr(source, 'read') and hasattr(source, 'mode'):
+                if 'U' in source.mode:
+                    # TemplateLoader of Genshi <= 0.6 open files with universal
+                    # newlines which is not suitable for zipfile
+                    raise ValueError('filepath is required '
+                        'if source is openned with universal newlines')
+                else:
+                    # source could be closed before generate calls
+                    source = BytesIO(source.read())
+        else:
+            source = self.filepath
+        self._zip_source = zf = zipfile.ZipFile(source)
         content = zf.read('content.xml')
         styles = zf.read('styles.xml')
 
@@ -268,7 +284,6 @@ class Template(MarkupTemplate):
             content_files.append((c_path, c_parsed))
             styles_files.append((s_path, s_parsed))
 
-        zf.close()
         parsed = []
         for fpath, fparsed in content_files + styles_files:
             parsed.append((genshi.core.PI, ('relatorio', fpath), None))
@@ -279,7 +294,7 @@ class Template(MarkupTemplate):
     def insert_directives(self, content):
         """adds the genshi directives, handle the images and the innerdocs.
         """
-        tree = lxml.etree.parse(StringIO(content))
+        tree = lxml.etree.parse(BytesIO(content))
         root = tree.getroot()
 
         # assign default/fake namespaces so that documents do not need to
@@ -306,7 +321,7 @@ class Template(MarkupTemplate):
         self._handle_images(tree)
         self._handle_innerdocs(tree)
         self._escape_values(tree)
-        return StringIO(lxml.etree.tostring(tree))
+        return BytesIO(lxml.etree.tostring(tree))
 
     def _invert_style(self, tree):
         "inverts the text:a and text:span"
@@ -692,7 +707,7 @@ class Template(MarkupTemplate):
 
     def generate(self, *args, **kwargs):
         "creates the RelatorioStream."
-        serializer = OOSerializer(self.filepath)
+        serializer = OOSerializer(self._zip_source)
         kwargs['__relatorio_make_href'] = ImageHref(serializer.outzip,
                                                     serializer.manifest,
                                                     kwargs)
@@ -752,13 +767,17 @@ class DuplicateColumnHeaders(object):
 class Manifest(object):
 
     def __init__(self, content):
-        self.tree = lxml.etree.parse(StringIO(content))
+        self.tree = lxml.etree.parse(BytesIO(content))
         self.root = self.tree.getroot()
         self.namespaces = self.root.nsmap
 
     def __str__(self):
-        return lxml.etree.tostring(self.tree, encoding='UTF-8',
-                                   xml_declaration=True)
+        val = lxml.etree.tostring(self.tree, encoding='UTF-8',
+                                  xml_declaration=True)
+        # In Python 3, val will be bytes
+        if not isinstance(val, str):
+            return str(val, 'utf-8')
+        return val
 
     def add_file_entry(self, path, mimetype=None):
         manifest_namespace = self.namespaces['manifest']
@@ -782,7 +801,7 @@ class Manifest(object):
 class Meta(object):
 
     def __init__(self, content):
-        self.tree = lxml.etree.parse(StringIO(content))
+        self.tree = lxml.etree.parse(BytesIO(content))
         root = self.tree.getroot()
         self.namespaces = root.nsmap
         path = '/office:document-meta/office:meta'
@@ -816,17 +835,21 @@ class Meta(object):
         self.remove('printed-by')
         self.remove('creator', 'dc')
         self.remove('date', 'dc')
-        return lxml.etree.tostring(self.tree, encoding='UTF-8',
-                                   xml_declaration=True)
+        val = lxml.etree.tostring(self.tree, encoding='UTF-8',
+                                  xml_declaration=True)
+        # In Python 3, val will be bytes
+        if not isinstance(val, str):
+            return str(val, 'utf-8')
+        return val
 
 
 class OOSerializer:
 
-    def __init__(self, oo_path):
-        self.inzip = zipfile.ZipFile(oo_path)
+    def __init__(self, inzip):
+        self.inzip = inzip
         self.manifest = Manifest(self.inzip.read(MANIFEST))
         self.meta = Meta(self.inzip.read(META))
-        self.new_oo = StringIO()
+        self.new_oo = BytesIO()
         self.outzip = zipfile.ZipFile(self.new_oo, 'w')
         self.xml_serializer = genshi.output.XMLSerializer()
 
@@ -850,7 +873,8 @@ class OOSerializer:
                 new_info = zipfile.ZipInfo(f_info.filename, now)
                 for attr in ('compress_type', 'flag_bits', 'create_system'):
                     setattr(new_info, attr, getattr(f_info, attr))
-                serialized_stream = output_encode(self.xml_serializer(stream))
+                serialized_stream = output_encode(self.xml_serializer(stream),
+                    encoding='utf-8')
                 self.outzip.writestr(new_info, serialized_stream)
             elif f_info.filename == MANIFEST:
                 manifest_info = f_info
@@ -863,7 +887,6 @@ class OOSerializer:
         self.manifest.remove_file_entry(THUMBNAILS + '/')
         if manifest_info:
             self.outzip.writestr(manifest_info, str(self.manifest))
-        self.inzip.close()
         self.outzip.close()
 
         return self.new_oo
